﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.EvolutionaryTracking;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Encodings.SymbolicExpressionTreeEncoding {
  /// <summary>
  /// A base class for operators that manipulate real-valued vectors.
  /// </summary>
  [Item("TracingSymbolicExpressionTreeManipulator", "A base class for operators that manipulate symbolic expression trees.")]
  [StorableClass]
  public abstract class TracingSymbolicExpressionTreeManipulator : TracingSymbolicExpressionTreeOperator, ISymbolicExpressionTreeManipulator {
    private const string SymbolicExpressionTreeParameterName = "SymbolicExpressionTree";

    #region Parameter Properties
    public ILookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ILookupParameter<ISymbolicExpressionTree>)Parameters[SymbolicExpressionTreeParameterName]; }
    }
    #endregion

    #region Properties
    public ISymbolicExpressionTree SymbolicExpressionTree {
      get { return SymbolicExpressionTreeParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    protected TracingSymbolicExpressionTreeManipulator(bool deserializing) : base(deserializing) { }
    protected TracingSymbolicExpressionTreeManipulator(TracingSymbolicExpressionTreeManipulator original, Cloner cloner) : base(original, cloner) { }
    public TracingSymbolicExpressionTreeManipulator()
      : base() {
      Parameters.Add(new LookupParameter<ISymbolicExpressionTree>(SymbolicExpressionTreeParameterName, "The symbolic expression tree on which the operator should be applied."));
    }

    public sealed override IOperation Apply() {
      ISymbolicExpressionTree tree = SymbolicExpressionTreeParameter.ActualValue;

      AddTracingVariablesToGlobalScope();

      /* The following block needs to be explained in more detail;
       * If the tree was already affected by crossover (before mutation), then:
       * - it will already be contained in the GlobalTraceMap
       * - In order to preserve information, a clone of the tree is created
       * - The clone represents the intermediate stage of the tree, before mutation is applied
       * - The clone therefore represents the child after crossover so GlobalTraceMap[clone] will get tree's parents
       *   and GlobalFragmentMap[clone] will get the subtree from clone that represents the cloned fragment
       * - After tree is mutated, we set GlobalTraceMap[tree] = clone and GlobalFragmentMap[tree] = modified node from tree
       */
      ISymbolicExpressionTreeNode fragment;
      if (GlobalTraceMap.ContainsKey(tree)) {
        var clone = (IItem)tree.Clone(); // create clone of tree
        GlobalTraceMap[clone] = GlobalTraceMap[tree]; // clone gets parents of tree
        GlobalTraceMap[tree] = new ItemList<IItem> { clone }; // tree gets clone as parent
        fragment = ((GenericWrapper<ISymbolicExpressionTreeNode>)GlobalFragmentMap[tree]).Content;
        int index = tree.IterateNodesBreadth().ToList().IndexOf(fragment); // returns -1 if fragment is not present in tree
        GlobalFragmentMap[clone] = new GenericWrapper<ISymbolicExpressionTreeNode>(index == -1 ? null : ((ISymbolicExpressionTree)clone).IterateNodesBreadth().ElementAt(index));
      } else {
        GlobalTraceMap[tree] = new ItemList<IItem> { GlobalCloneMap[tree] };
      }

      var original = GlobalCloneMap[tree];
      var nodes0 = ((ISymbolicExpressionTree)original).IterateNodesBreadth().ToList();

      // perform mutation
      Manipulate(RandomParameter.ActualValue, tree);

      var nodes1 = tree.IterateNodesBreadth().ToList();

      int min = Math.Min(nodes0.Count, nodes1.Count);
      // some tree manipulators only change local parameters while the actual node stays the same
      // consequently, we need a stronger comparison than ReferenceEquals
      var comp = SymbolicExpressionTreeNodeComparer;
      if (comp == null)
        throw new Exception(Name + ": Could not get SymbolicExpressionTreeNodeComparerParameter!");
      var mismatches = new List<int>();
      for (int i = 0; i != min; ++i) {
        if (!comp.Equals(nodes0[i], nodes1[i])) {
          mismatches.Add(i);
        }
      }
      if (mismatches.Count == 0) // should never happen
        fragment = null;
      else
        fragment = mismatches.Count > 1 ? nodes1[mismatches[0]].Parent : nodes1[mismatches[0]];
      GlobalFragmentMap[tree] = new GenericWrapper<ISymbolicExpressionTreeNode>(fragment);
      return base.Apply();
    }

    protected abstract void Manipulate(IRandom random, ISymbolicExpressionTree symbolicExpressionTree);
  }
}
