﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.DataAnalysis.Symbolic;
// type definitions for convenience
using HeuristicLab.Problems.DataAnalysis.Symbolic.Regression;
using CloneMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItem>;
using SymbolicExpressionTreeNodeItem = HeuristicLab.EvolutionaryTracking.GenericWrapper<HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.ISymbolicExpressionTreeNode>;
using TraceMapType = HeuristicLab.Core.ItemDictionary<HeuristicLab.Core.IItem, HeuristicLab.Core.IItemList<HeuristicLab.Core.IItem>>;

namespace HeuristicLab.EvolutionaryTracking {
  /// <summary>
  /// An operator that gathers information on tree fragments that get passed from one generation to the next via genetic operators
  /// (Tracking of genetic variance and heredity)
  /// </summary>
  [Item("SymbolicExpressionTreeFragmentsAnalyzer", "An operator that provides statistics about crossover fragments")]
  [StorableClass]
  public sealed class SymbolicExpressionTreeFragmentsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    #region Parameter names
    private const string UpdateIntervalParameterName = "UpdateInterval";
    private const string UpdateCounterParameterName = "UpdateCounter";
    private const string ResultsParameterName = "Results";
    private const string SecondaryTraceMapParameterName = "SecondaryTraceMap";
    private const string SecondaryFragmentMapParameterName = "SecondaryFragmentMap";
    private const string SecondaryCloneMapParameterName = "SecondaryCloneMap";
    private const string GlobalTraceMapParameterName = "GlobalTraceMap";
    private const string GlobalCloneMapParameterName = "GlobalCloneMap";
    private const string GlobalFragmentMapParameterName = "GlobalFragmentMap";
    private const string GenerationsParameterName = "Generations";
    private const string SymbolicExpressionInterpreterParameterName = "SymbolicExpressionTreeInterpreter";
    private const string SymbolicRegressionProblemDataParameterName = "ProblemData";
    private const string FragmentStatisticsParameterName = "FragmentStatistics";
    private const string FragmentLengthsDistributionParametersName = "FragmentLengths";
    private const string FragmentFrequenciesParameterName = "FragmentFrequencies";
    private const string ParentsDiversityParameterName = "ParentsDiversity";
    #endregion

    // impact values calculator
    private readonly SymbolicRegressionSolutionValuesCalculator calculator = new SymbolicRegressionSolutionValuesCalculator();

    #region Parameter properties
    public ValueParameter<IntValue> UpdateIntervalParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateIntervalParameterName]; }
    }
    public ValueParameter<IntValue> UpdateCounterParameter {
      get { return (ValueParameter<IntValue>)Parameters[UpdateCounterParameterName]; }
    }
    public LookupParameter<ResultCollection> ResultsParameter {
      get { return (LookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public LookupParameter<IntValue> GenerationsParameter {
      get { return (LookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    // tracking structures
    public LookupParameter<TraceMapType> GlobalTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[GlobalTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalCloneMapParameterName]; }
    }
    public LookupParameter<CloneMapType> GlobalFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[GlobalFragmentMapParameterName]; }
    }
    public LookupParameter<DataTable> FragmentFrequenciesParameter {
      get { return (LookupParameter<DataTable>)Parameters[FragmentFrequenciesParameterName]; }
    }
    public LookupParameter<DataTable> FragmentStatisticsParameter {
      get { return (LookupParameter<DataTable>)Parameters[FragmentStatisticsParameterName]; }
    }
    public LookupParameter<DataTable> FragmentLengthsParameter {
      get { return (LookupParameter<DataTable>)Parameters[FragmentLengthsDistributionParametersName]; }
    }
    public LookupParameter<DataTable> ParentsDiversityParameter {
      get { return (LookupParameter<DataTable>)Parameters[ParentsDiversityParameterName]; }
    }
    // auxiliary structures for tracking fragments and individuals from the previous generation
    // (this is needed because tracking is done in connection to the current generation, i.e., for 
    // counting "successful" offspring and fragments
    public LookupParameter<TraceMapType> SecondaryTraceMapParameter {
      get { return (LookupParameter<TraceMapType>)Parameters[SecondaryTraceMapParameterName]; }
    }
    public LookupParameter<CloneMapType> SecondaryFragmentMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[SecondaryFragmentMapParameterName]; }
    }
    public LookupParameter<CloneMapType> SecondaryCloneMapParameter {
      get { return (LookupParameter<CloneMapType>)Parameters[SecondaryCloneMapParameterName]; }
    }
    // problem data, interpreter and evaluator
    public LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter> SymbolicExpressionInterpreterParameter {
      get { return (LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter>)Parameters[SymbolicExpressionInterpreterParameterName]; }
    }
    public LookupParameter<RegressionProblemData> SymbolicRegressionProblemDataParameter {
      get { return (LookupParameter<RegressionProblemData>)Parameters[SymbolicRegressionProblemDataParameterName]; }
    }
    #endregion

    #region Parameters
    public bool EnabledByDefault {
      get { return true; }
    }
    public IntValue UpdateInterval {
      get { return UpdateIntervalParameter.Value; }
    }
    public IntValue UpdateCounter {
      get { return UpdateCounterParameter.Value; }
    }
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public CloneMapType GlobalCloneMap {
      get { return GlobalCloneMapParameter.ActualValue; }
    }
    public TraceMapType GlobalTraceMap {
      get { return GlobalTraceMapParameter.ActualValue; }
    }
    public TraceMapType SecondaryTraceMap {
      get { return SecondaryTraceMapParameter.ActualValue; }
      set { SecondaryTraceMapParameter.ActualValue = value; }
    }
    public CloneMapType SecondaryFragmentMap {
      get { return SecondaryFragmentMapParameter.ActualValue; }
      set { SecondaryFragmentMapParameter.ActualValue = value; }
    }
    public CloneMapType SecondaryCloneMap {
      get { return SecondaryCloneMapParameter.ActualValue; }
      set { SecondaryCloneMapParameter.ActualValue = value; }
    }
    public CloneMapType GlobalFragmentMap {
      get { return GlobalFragmentMapParameter.ActualValue; }
    }
    public IntValue Generations {
      get { return GenerationsParameter.ActualValue; }
    }
    public DataTable FragmentStatistics {
      get { return FragmentStatisticsParameter.ActualValue; }
    }
    public DataTable FragmentLengths {
      get { return FragmentLengthsParameter.ActualValue; }
    }
    public DataTable FragmentFrequencies {
      get { return FragmentFrequenciesParameter.ActualValue; }
    }
    public DataTable ParentsDiversity {
      get { return ParentsDiversityParameter.ActualValue; }
    }
    public SymbolicDataAnalysisExpressionTreeInterpreter SymbolicExpressionInterpreter {
      get { return SymbolicExpressionInterpreterParameter.ActualValue; }
    }
    public RegressionProblemData SymbolicRegressionProblemData {
      get { return SymbolicRegressionProblemDataParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private SymbolicExpressionTreeFragmentsAnalyzer(bool deserializing) : base(deserializing) { }
    private SymbolicExpressionTreeFragmentsAnalyzer(SymbolicExpressionTreeFragmentsAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicExpressionTreeFragmentsAnalyzer(this, cloner);
    }
    public SymbolicExpressionTreeFragmentsAnalyzer()
      : base() {
      #region Add parameters
      // analyzer update counter and update interval
      Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
      // tracking
      Parameters.Add(new LookupParameter<TraceMapType>(GlobalTraceMapParameterName, "A global cache containing the whole genealogy."));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalCloneMapParameterName, "A global map keeping track of trees and their clones (made during selection)."));
      Parameters.Add(new LookupParameter<CloneMapType>(GlobalFragmentMapParameterName, "A global map keeping track of tree fragments received via crossover."));
      // secondary tracking 
      Parameters.Add(new LookupParameter<TraceMapType>(SecondaryTraceMapParameterName, "Parameter to store the trace map from the previous generation"));
      Parameters.Add(new LookupParameter<CloneMapType>(SecondaryCloneMapParameterName, "Parameter to store the clone map from the previous generation"));
      Parameters.Add(new LookupParameter<CloneMapType>(SecondaryFragmentMapParameterName, "Parameter to store the fragment map from the previous generation"));
      // fragment statistics parameters
      Parameters.Add(new ValueLookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations so far."));
      Parameters.Add(new LookupParameter<DataTable>(FragmentStatisticsParameterName, "The data table to store the fragment statistics."));
      Parameters.Add(new LookupParameter<DataTable>(FragmentLengthsDistributionParametersName, "The data table to store the distribution of fragment lengths."));
      Parameters.Add(new LookupParameter<DataTable>(FragmentFrequenciesParameterName, "A data structure for fragment frequencies"));
      Parameters.Add(new LookupParameter<DataTable>(ParentsDiversityParameterName, "A data structure to store the diversity of the parents per generation."));
      // impact calculation
      Parameters.Add(new LookupParameter<SymbolicDataAnalysisExpressionTreeInterpreter>(SymbolicExpressionInterpreterParameterName, "Interpreter for symbolic expression trees"));
      Parameters.Add(new LookupParameter<RegressionProblemData>(SymbolicRegressionProblemDataParameterName, "The symbolic data analysis problem."));
      #endregion

      UpdateCounterParameter.Hidden = true;
      UpdateIntervalParameter.Hidden = true;
    }
    #region After deserialization code
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      // check if all the parameters are present and accounted for 
      if (!Parameters.ContainsKey(UpdateIntervalParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateIntervalParameterName, "The interval in which the tree length analysis should be applied.", new IntValue(1)));
      }
      if (!Parameters.ContainsKey(UpdateCounterParameterName)) {
        Parameters.Add(new ValueParameter<IntValue>(UpdateCounterParameterName, "The value which counts how many times the operator was called since the last update", new IntValue(0)));
        UpdateCounterParameter.Hidden = true;
      }
      if (!Parameters.ContainsKey(FragmentStatisticsParameterName)) {
        Parameters.Add(new LookupParameter<DataTable>(FragmentStatisticsParameterName, "The data table to store the fragment statistics."));
      }
      if (!Parameters.ContainsKey(FragmentLengthsDistributionParametersName)) {
        Parameters.Add(new LookupParameter<DataTable>(FragmentLengthsDistributionParametersName, "The data table to store the distribution of fragment lengths."));
      }
      if (!Parameters.ContainsKey(FragmentFrequenciesParameterName)) {
        Parameters.Add(new LookupParameter<ItemList<ItemList<IItem>>>(FragmentFrequenciesParameterName, "A data structure for fragment frequencies"));
      }
      if (!Parameters.ContainsKey(ParentsDiversityParameterName)) {
        Parameters.Add(new LookupParameter<DataTable>(ParentsDiversityParameterName, "A data table to store the diversity of the parents per generation."));
      }
    }
    #endregion
    #region IStatefulItem members
    public override void InitializeState() {
      base.InitializeState();
      UpdateCounter.Value = 0;
    }

    public override void ClearState() {
      base.ClearState();
      UpdateCounter.Value = 0;
    }
    #endregion

    /* A small description of the way this analyzer works
     * 
     * We keep 3 maps in the global scope: the GlobalTraceMap, the GlobalFragmentMap and the GlobalCloneMap that are updated on each step: selection, crossover, mutation
     * These maps provide information about the children, parents and transferred fragments as follows:
     * 
     * 1) GlobalCloneMap
     *    This data structure provides the basis for tracking. In the selection step, every selected individual gets cloned and transferred into the recombination pool.
     *    The GlobalCloneMap keeps track of clones by mapping them to the original individuals. This is an essential step for building the genealogy graph. Using this
     *    structure we can find out how many times each individual was selected. 
     *    The GlobalCloneMap consists of Key-Value pairs Clone->Original
     *
     * 2) GlobalTraceMap
     *    Keeps information in the form of key-value pairs Child->{Parents}, where {Parents} is a set consisting of 
     *      - one parent individual (in the case of mutation), by language abuse we say that the individual that mutation acted on is the parent,
     *        and the resulting individual is the child
     *      - two parent individuals (in the case of crossover), named Parent0 and Parent1. Parent0 is the individual that crossover acts on by replacing 
     *        one of its subtrees with a compatible subtree, randomly selected from Parent1
     *
     *    CROSSOVER
     *        Parent0     Parent1                                                                    {Parents}
     *           \           /                                                                           ^
     *            \       fragment (inserted from Parent1 into Parent0 to create the Child)              ^    [Mapping provided by the GlobalTraceMap]
     *             \       /                                                                             ^
     *              \     /                                                                              ^
     *               Child                                                                             Child
     *               
     *    MUTATION
     *        Parent0
     *           |
     *           |    [mutation acts on a random node/subtree in the 'parent' individual
     *           |
     *         Child
     *         
     *    Since crossover and mutation act on individuals in the recombination pool (which in fact are clones of individuals from the previous generation), the GlobalTraceMap
     *    is populated with values given by the mappings in the GlobalCloneMap, so that the original parent for each child is known.
     *    
     *  3) GlobalFragmentMap
     *     Keeps information in the form of Key-Value pairs about an individual and its respective fragment 
     *     (ie., the subtree received via crossover or created/altered by mutation)
     * */
    public override IOperation Apply() {
      UpdateCounter.Value++;
      if (UpdateCounter.Value == UpdateInterval.Value) {
        UpdateCounter.Value = 0; // reset counter
        if (Generations.Value > 0) {
          InitializeParameters();

          CalculateFragmentStatistics();

          // save the current generation so it can be compared to the following one, next time the analyzer is applied
          SecondaryTraceMap = (TraceMapType)GlobalTraceMap.Clone();
          SecondaryTraceMap.Clear();
          foreach (var m in GlobalTraceMap)
            SecondaryTraceMap.Add(m.Key, m.Value);

          SecondaryCloneMap.Clear();
          foreach (var m in GlobalCloneMap)
            SecondaryCloneMap.Add(m.Key, m.Value);

          SecondaryFragmentMap.Clear();
          foreach (var m in GlobalFragmentMap)
            SecondaryFragmentMap.Add(m.Key, m.Value);
        }
      }
      if (Generations.Value > 0)
        if (this.Successor == null || !(this.Successor is SymbolicExpressionTreeGenealogyAnalyzer)) {
          // clear the global maps to save memory
          GlobalCloneMap.Clear();
          GlobalTraceMap.Clear();
          GlobalFragmentMap.Clear();
        }
      return base.Apply();
    }

    private void InitializeParameters() {
      #region Init parameters
      var results = ResultsParameter.ActualValue;

      if (FragmentStatistics == null) {
        FragmentStatisticsParameter.ActualValue = new DataTable("Fragment Statistics", "Statistical measurements of fragments aggregated over the whole population");
        FragmentStatistics.VisualProperties.YAxisTitle = "Fragment length/Similarities";
        results.Add(new Result("Fragment Statistics", FragmentStatistics));
      }
      if (FragmentLengths == null) {
        FragmentLengthsParameter.ActualValue = new DataTable("Fragment Lengths", "Histograms for the distribution of fragment and individual lengths");
        FragmentLengths.VisualProperties.YAxisTitle = "Frequency";
        results.Add(new Result("Fragment Lengths", FragmentLengths));
      }
      if (FragmentFrequencies == null) {
        FragmentFrequenciesParameter.ActualValue = new DataTable("Fragment Frequencies", "Frequencies of good and high impact fragments");
        FragmentFrequencies.VisualProperties.YAxisTitle = "Frequency";
        results.Add(new Result("Fragment Frequencies", FragmentFrequencies));
      }
      if (ParentsDiversity == null) {
        ParentsDiversityParameter.ActualValue = new DataTable("Parents Diversity", "Number of unique parents per generation");
        ParentsDiversity.VisualProperties.YAxisTitle = "Unique parents";
        results.Add(new Result("Parents Diversity", ParentsDiversity));
      }
      if (SecondaryTraceMap == null) {
        SecondaryTraceMap = new TraceMapType();
      }
      if (SecondaryCloneMap == null) {
        SecondaryCloneMap = new CloneMapType();
      }
      if (SecondaryFragmentMap == null) {
        SecondaryFragmentMap = new CloneMapType();
      }
      #endregion
    }

    private void InitializeRows() {
      if (!FragmentStatistics.Rows.ContainsKey("Parent lengths (good)")) {
        FragmentStatistics.Rows.Add(new DataRow("Parent lengths (good)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Parent lengths (all)")) {
        FragmentStatistics.Rows.Add(new DataRow("Parent lengths (all)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Child lengths (good)")) {
        FragmentStatistics.Rows.Add(new DataRow("Child lengths (good)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Child lengths (all)")) {
        FragmentStatistics.Rows.Add(new DataRow("Child lengths (all)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Fragment lengths (good)")) {
        FragmentStatistics.Rows.Add(new DataRow("Fragment lengths (good)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Fragment lengths (all)")) {
        FragmentStatistics.Rows.Add(new DataRow("Fragment lengths (all)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Fragment lengths (high impact)")) {
        FragmentStatistics.Rows.Add(new DataRow("Fragment lengths (high impact)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Avg visitation length (all children)")) {
        FragmentStatistics.Rows.Add(new DataRow("Avg visitation length (all children)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Avg visitation length (good children)")) {
        FragmentStatistics.Rows.Add(new DataRow("Avg visitation length (good children)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      // exact similarity
      if (!FragmentStatistics.Rows.ContainsKey("Exact similarity (good fragments)")) {
        FragmentStatistics.Rows.Add(new DataRow("Exact similarity (good fragments)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentStatistics.Rows.ContainsKey("Exact similarity (all fragments)")) {
        FragmentStatistics.Rows.Add(new DataRow("Exact similarity (all fragments)") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentFrequencies.Rows.ContainsKey("Frequency of useful fragments")) {
        FragmentFrequencies.Rows.Add(new DataRow("Frequency of useful fragments") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!FragmentFrequencies.Rows.ContainsKey("Frequency of high impact fragments")) {
        FragmentFrequencies.Rows.Add(new DataRow("Frequency of high impact fragments") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      if (!ParentsDiversity.Rows.ContainsKey("Unique parents")) {
        ParentsDiversity.Rows.Add(new DataRow("Unique parents") {
          VisualProperties = { StartIndexZero = true }
        });
      }
      double scaleFactor = 1.0;
      if (!FragmentLengths.Rows.ContainsKey("All fragments length distribution"))
        FragmentLengths.Rows.Add(new DataRow("All fragments length distribution") {
          VisualProperties = {
            StartIndexZero = true,
            ChartType = DataRowVisualProperties.DataRowChartType.Histogram,
            ScaleFactor = scaleFactor
          }
        });
      if (!FragmentLengths.Rows.ContainsKey("Useful fragments length distribution"))
        FragmentLengths.Rows.Add(new DataRow("Useful fragments length distribution") {
          VisualProperties = {
            StartIndexZero = true,
            ChartType = DataRowVisualProperties.DataRowChartType.Histogram,
            ScaleFactor = scaleFactor
          }
        });
      if (!FragmentLengths.Rows.ContainsKey("All children length distribution"))
        FragmentLengths.Rows.Add(new DataRow("All children length distribution") {
          VisualProperties = {
            StartIndexZero = true,
            ChartType = DataRowVisualProperties.DataRowChartType.Histogram,
            ScaleFactor = scaleFactor
          }
        });
      if (!FragmentLengths.Rows.ContainsKey("Useful children length distribution"))
        FragmentLengths.Rows.Add(new DataRow("Useful children length distribution") {
          VisualProperties = {
            StartIndexZero = true,
            ChartType = DataRowVisualProperties.DataRowChartType.Histogram,
            ScaleFactor = scaleFactor
          }
        });
    }

    private void CalculateFragmentStatistics() {
      // for this kind of tracking/analysis, we need at least 2 successive generations of crossover
      if (Generations.Value > 1) {
        InitializeRows();

        #region Fragment Statistics
        // create a hashset for fast lookup of parents (to see which children from the previous generation become parents, and count their fragments)
        var parentsLookup = new HashSet<ISymbolicExpressionTree>(GlobalTraceMap.Values.SelectMany(x => x.Cast<ISymbolicExpressionTree>()));
        var allParents = SecondaryTraceMap.Values.SelectMany(x => x.Cast<ISymbolicExpressionTree>()).ToList();
        var allChildren = SecondaryFragmentMap.Keys.Cast<ISymbolicExpressionTree>().ToList(); /* SecondaryTraceMap.Keys should be the same with SecondaryFragmentMap.Keys */
        var allFragments = SecondaryFragmentMap.Values.Cast<SymbolicExpressionTreeNodeItem>().Select(x => x.Content).Where(x => x != null).ToList();
        // "good" fragments are fragments contained in a surviving offspring (that became a parent)
        // "good" parents are those that produced a surviving offspring
        // "good" children are the surviving offspring
        var goodFragments = new List<ISymbolicExpressionTreeNode>();
        var goodParents = new List<ISymbolicExpressionTree>();
        var goodChildren = new List<ISymbolicExpressionTree>();

        // take all individuals that have received a fragment in the previous generation
        foreach (var m in SecondaryFragmentMap) {
          var individual = m.Key as ISymbolicExpressionTree;
          // check if the individual became a parent in the current generation,
          // by checking if it appears among the parents from the current generation trace map
          if (parentsLookup.Contains(individual)) {
            var symbExprTreeNodeItem = m.Value as SymbolicExpressionTreeNodeItem;
            var fragment = symbExprTreeNodeItem.Content;
            if (fragment != null) {
              goodFragments.Add(fragment);
              goodParents.AddRange(SecondaryTraceMap[individual].Cast<ISymbolicExpressionTree>());
              goodChildren.Add(individual);
            }
          }
        }

        if (false) {
          var highImpactFragments = new List<ISymbolicExpressionTreeNode>();
          foreach (var child in goodChildren) {
            var impactValues = calculator.CalculateImpactValues(child, SymbolicExpressionInterpreter, SymbolicRegressionProblemData, 0, 0);
            var fragmentItem = SecondaryFragmentMap[child] as SymbolicExpressionTreeNodeItem;
            var fragment = fragmentItem.Content;
            if (fragment != null && impactValues[fragment] > 0.1) {
              if (highImpactFragments.Contains(fragment, new SymbolicExpressionTreeNodeSimilarityComparer(SimilarityLevel.Exact)))
                continue;
              // prune fragment (remove irrelevant/low impact nodes)
              // this works because a child will never have an impact value greater than its parent
              foreach (var fnode in fragment.IterateNodesBreadth().Where(x => x.SubtreeCount > 0)) {
                for (int i = 0; i < fnode.SubtreeCount; ++i) {
                  var subtree = fnode.GetSubtree(i);
                  if (impactValues[subtree] < 0.1)
                    fnode.RemoveSubtree(i);
                }
              }
              highImpactFragments.Add(fragment);
            }
          }
        }

        FragmentFrequencies.Rows["Frequency of useful fragments"].Values.Add(goodFragments.Count);
        //        FragmentFrequencies.Rows["Frequency of high impact fragments"].Values.Add(highImpactFragments.Count);
        FragmentStatistics.Rows["Fragment lengths (good)"].Values.Add(goodFragments.Average(x => x.GetLength()));
        FragmentStatistics.Rows["Fragment lengths (all)"].Values.Add(allFragments.Average(x => x.GetLength()));
        //        double avg = highImpactFragments.Count > 0 ? highImpactFragments.Average(x => x.GetLength()) : 0;
        //        FragmentStatistics.Rows["Fragment lengths (high impact)"].Values.Add(avg);
        FragmentStatistics.Rows["Parent lengths (good)"].Values.Add(goodParents.Average(x => x.Length));
        FragmentStatistics.Rows["Parent lengths (all)"].Values.Add(allParents.Average(x => x.Length));
        FragmentStatistics.Rows["Child lengths (good)"].Values.Add(goodChildren.Average(x => x.Length));
        FragmentStatistics.Rows["Child lengths (all)"].Values.Add(allChildren.Average(x => x.Length));
        //        FragmentStatistics.Rows["Avg visitation length (all children)"].Values.Add(allChildren.Average(x => VisitationLength(x)));
        //        FragmentStatistics.Rows["Avg visitation length (good children)"].Values.Add(goodChildren.Average(x => VisitationLength(x)));
        FragmentLengths.Rows["All fragments length distribution"].Values.Replace(allFragments.Select(x => (double)x.GetLength()));
        FragmentLengths.Rows["Useful fragments length distribution"].Values.Replace(goodFragments.Select(x => (double)x.GetLength()));
        FragmentLengths.Rows["All children length distribution"].Values.Replace(allChildren.Select(x => (double)x.Length));
        FragmentLengths.Rows["Useful children length distribution"].Values.Replace(goodChildren.Select(x => (double)x.Length));
        ParentsDiversity.Rows["Unique parents"].Values.Add(SecondaryCloneMap.Values.GroupBy(x => x).Count());
        #endregion
      }
    }

    private static int VisitationLength(ISymbolicExpressionTree tree) {
      return VisitationLength(tree.Root);
    }

    private static int VisitationLength(ISymbolicExpressionTreeNode node) {
      return node.IterateNodesBreadth().Sum(n => n.GetLength());
    }

    #region Similarity computations
    /// <summary>
    /// Provide a measure of how similar the fragments that get passed by crossover from one generation to the next are. 
    /// </summary>
    /// <param name="fragments">The symbolic expression tree fragments</param>
    /// <param name="mode">The similarity mode (0 - exact, 1 - high, 2 - relaxed)</param>
    /// <returns>The average number of similar fragments</returns>
    private static double CalculateSimilarity(IList<ISymbolicExpressionTreeNode> fragments, SimilarityLevel similarity) {
      var visited = new bool[fragments.Count];
      int groups = 0;
      for (int i = 0; i != fragments.Count - 1; ++i) {
        if (visited[i]) continue;
        for (int j = i + 1; j != fragments.Count; ++j) {
          if (visited[j]) continue;
          if (fragments[i].IsSimilarTo(fragments[j], similarity))
            visited[j] = true;
        }
        ++groups;
      }
      return (double)fragments.Count / groups;
    }
    #endregion
  } //SymbolicExpressionTreeFragmentsAnalyzer
}
